94057c3d3e
PR Test (NPU) / check-changes (push) Has been cancelled
PR Test (NPU) / pr-gate (push) Has been cancelled
PR Test (NPU) / set-image-config (push) Has been cancelled
PR Test (NPU) / stage-b-test-1-npu-a2 (0) (push) Has been cancelled
PR Test (NPU) / stage-b-test-1-npu-a2 (1) (push) Has been cancelled
PR Test (NPU) / stage-b-test-2-npu-a2 (0) (push) Has been cancelled
PR Test (NPU) / stage-b-test-2-npu-a2 (1) (push) Has been cancelled
PR Test (NPU) / stage-b-test-4-npu-a3 (push) Has been cancelled
PR Test (NPU) / stage-b-test-16-npu-a3 (push) Has been cancelled
PR Test (NPU) / multimodal-gen-test-1-npu-a3 (push) Has been cancelled
PR Test (NPU) / multimodal-gen-test-2-npu-a3 (push) Has been cancelled
PR Test (Arm64) / pr-gate (push) Has been cancelled
PR Test (Arm64) / check-changes (push) Has been cancelled
PR Test (Arm64) / build-test (push) Has been cancelled
PR Test (sgl-router) / gate (push) Has been cancelled
PR Test (sgl-router) / tier-1 — lint (push) Has been cancelled
PR Test (sgl-router) / tier-2 — build + test (push) Has been cancelled
PR Test (sgl-router) / tier-3 — docker (placeholder) (push) Has been cancelled
PR Test (sgl-router) / tier-3 — k8s integration (push) Has been cancelled
PR Test (sgl-router) / tier-3 — e2e (push) Has been cancelled
PR Test (sgl-router) / finish (push) Has been cancelled
PR Test (NPU) / single-node-poc (map[name:qwen3_6_27b_w8a8_1p_in64k_out1k_50ms runner:linux-aarch64-a3-2 test_case:test/registered/ascend/performance/qwen3_6_27b/test_npu_qwen3_6_27b_w8a8_1p_in64k_out1k_50ms.py test_type:perf]) (push) Has been cancelled
PR Test (NPU) / pr-test-npu-finish (push) Has been cancelled
PR Test (Xeon) / pr-gate (push) Has been cancelled
PR Test (Xeon) / check-changes (push) Has been cancelled
PR Test (Xeon) / build-test (, xeon-gnr, base-b-test-cpu) (push) Has been cancelled
PR Test (XPU) / check-changes (push) Has been cancelled
PR Test (XPU) / pr-gate (push) Has been cancelled
PR Test (XPU) / stage-a-test-1-gpu-xpu (push) Has been cancelled
PR Test (XPU) / wait-for-stage-a (push) Has been cancelled
PR Test (XPU) / stage-b-test-1-gpu-xpu (push) Has been cancelled
PR Test (XPU) / finish (push) Has been cancelled
CI Model Inventory / build-inventory (push) Has been cancelled
Lint / lint (push) Has been cancelled
PR Benchmark (SMG Components) / Benchmark Compilation Check (push) Has been cancelled
PR Benchmark (SMG Components) / Benchmark - Manual Policy (push) Has been cancelled
PR Benchmark (SMG Components) / Benchmark - Request Processing (push) Has been cancelled
PR Benchmark (SMG Components) / Benchmark Summary (push) Has been cancelled
PR Test (SMG) / build-wheel (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on windows (x86_64 - auto) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on macos (x86_64 - auto) (push) Has been cancelled
PR Test (SMG) / python-unit-tests (push) Has been cancelled
PR Test (SMG) / unit-tests (push) Has been cancelled
PR Test (SMG) / benchmarks (push) Has been cancelled
PR Test (SMG) / chat-completions (push) Has been cancelled
PR Test (SMG) / chat-completions-4gpu (push) Has been cancelled
PR Test (SMG) / e2e (push) Has been cancelled
PR Test (SMG) / docker-build-test (push) Has been cancelled
PR Test (SMG) / k8s-integration (push) Has been cancelled
PR Test (SMG) / finish (push) Has been cancelled
PR Test (SMG) / summarize-benchmarks (push) Has been cancelled
Release SGLang Model Gateway Docker Image / publish (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on macos (aarch64 - auto) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on linux (aarch64 - auto) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on linux (x86_64 - auto) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on linux (aarch64 - musllinux_1_1) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on linux (x86_64 - musllinux_1_1) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / Build SDist (push) Has been cancelled
Release SGLang Model Gateway to PyPI / Upload to PyPI (push) Has been cancelled
Release SGLang Kernels / build-cu129-matrix (aarch64, 12.9, 3.10, arm-kernel-build-node) (push) Has been cancelled
Release SGLang Kernels / build-cu129-matrix (x86_64, 12.9, 3.10, x64-kernel-build-node) (push) Has been cancelled
Release SGLang Kernels / release-cu129 (push) Has been cancelled
Release SGLang Kernels / build-cu130-matrix (aarch64, 13.0, 3.10, arm-kernel-build-node) (push) Has been cancelled
Release SGLang Kernels / build-cu130-matrix (x86_64, 13.0, 3.10, x64-kernel-build-node) (push) Has been cancelled
Release SGLang Kernels / release-cu130 (push) Has been cancelled
Release SGLang Kernels / build-rocm-matrix (3.10, 700) (push) Has been cancelled
Release SGLang Kernels / build-rocm-matrix (3.10, 720) (push) Has been cancelled
Release SGLang Kernels / release-rocm700 (push) Has been cancelled
Release SGLang Kernels / release-rocm720 (push) Has been cancelled
Release SGLang Kernels / build-musa43 (43, 3.10) (push) Has been cancelled
Release SGLang Kernels / release-musa43 (push) Has been cancelled
175 lines
4.8 KiB
Python
175 lines
4.8 KiB
Python
# Adapted from https://github.com/fla-org/flash-linear-attention/blob/main/fla/ops/common/chunk_o.py
|
|
# -*- coding: utf-8 -*-
|
|
# Copyright (c) 2023-2025, Songlin Yang, Yu Zhang
|
|
|
|
from typing import Optional
|
|
|
|
import torch
|
|
import triton
|
|
import triton.language as tl
|
|
|
|
from sglang.srt.layers.attention.fla.index import prepare_chunk_indices
|
|
from sglang.srt.layers.attention.fla.op import exp, safe_exp
|
|
from sglang.srt.layers.attention.fla.utils import check_shared_mem, is_nvidia_hopper
|
|
|
|
BKV_LIST = [64, 128] if check_shared_mem() else [32, 64]
|
|
NUM_WARPS = [2, 4] if is_nvidia_hopper else [2, 4, 8]
|
|
|
|
|
|
# @triton.autotune(
|
|
# configs=[
|
|
# triton.Config({"BK": BK, "BV": BV}, num_warps=num_warps, num_stages=num_stages)
|
|
# for BK in BKV_LIST
|
|
# for BV in BKV_LIST
|
|
# for num_warps in NUM_WARPS
|
|
# for num_stages in [2, 3, 4]
|
|
# ],
|
|
# key=["H", "K", "V", "BT"],
|
|
# )
|
|
@triton.jit(do_not_specialize=["T"])
|
|
def chunk_fwd_kernel_o(
|
|
q,
|
|
k,
|
|
v,
|
|
h,
|
|
g,
|
|
o,
|
|
cu_seqlens,
|
|
chunk_indices,
|
|
scale,
|
|
T,
|
|
H: tl.constexpr,
|
|
Hg: tl.constexpr,
|
|
K: tl.constexpr,
|
|
V: tl.constexpr,
|
|
BT: tl.constexpr,
|
|
BK: tl.constexpr,
|
|
BV: tl.constexpr,
|
|
USE_G: tl.constexpr,
|
|
IS_VARLEN: tl.constexpr,
|
|
):
|
|
i_v, i_t, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
|
|
i_b, i_h = i_bh // H, i_bh % H
|
|
|
|
if IS_VARLEN:
|
|
i_tg = i_t
|
|
i_n, i_t = tl.load(chunk_indices + i_t * 2).to(tl.int32), tl.load(
|
|
chunk_indices + i_t * 2 + 1
|
|
).to(tl.int32)
|
|
bos, eos = tl.load(cu_seqlens + i_n).to(tl.int32), tl.load(
|
|
cu_seqlens + i_n + 1
|
|
).to(tl.int32)
|
|
T = eos - bos
|
|
NT = tl.cdiv(T, BT)
|
|
else:
|
|
NT = tl.cdiv(T, BT)
|
|
i_tg = i_b * NT + i_t
|
|
bos, eos = i_b * T, i_b * T + T
|
|
|
|
# offset calculation
|
|
q += (bos * Hg + i_h // (H // Hg)) * K
|
|
k += (bos * Hg + i_h // (H // Hg)) * K
|
|
v += (bos * H + i_h) * V
|
|
o += (bos * H + i_h) * V
|
|
h += (i_tg * H + i_h).to(tl.int64) * V * K
|
|
|
|
b_o = tl.zeros([BT, BV], dtype=tl.float32)
|
|
b_A = tl.zeros([BT, BT], dtype=tl.float32)
|
|
|
|
for i_k in range(tl.cdiv(K, BK)):
|
|
p_q = tl.make_block_ptr(
|
|
q, (T, K), (Hg * K, 1), (i_t * BT, i_k * BK), (BT, BK), (1, 0)
|
|
)
|
|
p_k = tl.make_block_ptr(
|
|
k, (K, T), (1, Hg * K), (i_k * BK, i_t * BT), (BK, BT), (0, 1)
|
|
)
|
|
p_h = tl.make_block_ptr(
|
|
h, (V, K), (K, 1), (i_v * BV, i_k * BK), (BV, BK), (1, 0)
|
|
)
|
|
# [BT, BK]
|
|
b_q = tl.load(p_q, boundary_check=(0, 1))
|
|
# [BK, BT]
|
|
b_k = tl.load(p_k, boundary_check=(0, 1))
|
|
# [BV, BK]
|
|
b_h = tl.load(p_h, boundary_check=(0, 1))
|
|
|
|
# [BT, BK] @ [BK, BV] -> [BT, BV]
|
|
b_o += tl.dot(b_q, tl.trans(b_h))
|
|
# [BT, BK] @ [BK, BT] -> [BT, BT]
|
|
b_A += tl.dot(b_q, b_k)
|
|
|
|
if USE_G:
|
|
g += bos * H + i_h
|
|
p_g = tl.make_block_ptr(g, (T,), (H,), (i_t * BT,), (BT,), (0,))
|
|
b_g = tl.load(p_g, boundary_check=(0,))
|
|
b_o = b_o * exp(b_g)[:, None]
|
|
b_A = b_A * safe_exp(b_g[:, None] - b_g[None, :])
|
|
|
|
o_i = tl.arange(0, BT)
|
|
m_A = o_i[:, None] >= o_i[None, :]
|
|
b_A = tl.where(m_A, b_A, 0)
|
|
|
|
p_v = tl.make_block_ptr(
|
|
v, (T, V), (H * V, 1), (i_t * BT, i_v * BV), (BT, BV), (1, 0)
|
|
)
|
|
p_o = tl.make_block_ptr(
|
|
o, (T, V), (H * V, 1), (i_t * BT, i_v * BV), (BT, BV), (1, 0)
|
|
)
|
|
b_v = tl.load(p_v, boundary_check=(0, 1))
|
|
|
|
# to fix mma -> mma layout conversion
|
|
# already solved by triton v3.2 or higher
|
|
b_o = b_o * scale + tl.dot(b_A.to(b_v.dtype), b_v) * scale
|
|
tl.store(p_o, b_o.to(p_o.dtype.element_ty), boundary_check=(0, 1))
|
|
|
|
|
|
def chunk_fwd_o(
|
|
q: torch.Tensor,
|
|
k: torch.Tensor,
|
|
v: torch.Tensor,
|
|
h: torch.Tensor,
|
|
g: Optional[torch.Tensor] = None, # cumsum of log decay
|
|
scale: Optional[float] = None,
|
|
cu_seqlens: Optional[torch.LongTensor] = None,
|
|
chunk_size: int = 64,
|
|
) -> torch.Tensor:
|
|
B, T, Hg, K, V = *q.shape, v.shape[-1]
|
|
H = v.shape[-2]
|
|
BT = min(chunk_size, max(16, triton.next_power_of_2(T)))
|
|
chunk_indices = (
|
|
prepare_chunk_indices(cu_seqlens, BT) if cu_seqlens is not None else None
|
|
)
|
|
NT = triton.cdiv(T, BT) if cu_seqlens is None else len(chunk_indices)
|
|
if scale is None:
|
|
scale = k.shape[-1] ** -0.5
|
|
|
|
o = torch.zeros_like(v)
|
|
|
|
def grid(meta):
|
|
return (triton.cdiv(V, meta["BV"]), NT, B * H)
|
|
|
|
chunk_fwd_kernel_o[grid](
|
|
q,
|
|
k,
|
|
v,
|
|
h,
|
|
g,
|
|
o,
|
|
cu_seqlens,
|
|
chunk_indices,
|
|
scale,
|
|
T=T,
|
|
H=H,
|
|
Hg=Hg,
|
|
K=K,
|
|
V=V,
|
|
BT=BT,
|
|
BK=128,
|
|
BV=64,
|
|
USE_G=g is not None,
|
|
IS_VARLEN=cu_seqlens is not None,
|
|
num_warps=4,
|
|
num_stages=2,
|
|
)
|
|
return o
|